"""
Relative Dynamic Causal Strength (rDCS) implementation.

This module provides the implementation of time-varying causality
including Transfer Entropy (TE), Dynamic Causal Strength (DCS),
and Relative Dynamic Causal Strength (rDCS).
"""

import logging
from typing import Dict

import numpy as np

from trancit.core.base import BaseAnalyzer, BaseResult
from trancit.core.exceptions import ComputationError, ValidationError
from trancit.utils.preprocess import regularize_if_singular

logger = logging.getLogger(__name__)


class RelativeDCSResult(BaseResult):
    """Result container for Relative Dynamic Causal Strength analysis."""

    def __init__(
        self,
        transfer_entropy: np.ndarray,
        dynamic_causal_strength: np.ndarray,
        relative_dynamic_causal_strength: np.ndarray,
        coefficients: np.ndarray,
    ):
        """
        Initialize Relative DCS result.

        Parameters
        ----------
        transfer_entropy : np.ndarray
            Transfer Entropy values
        dynamic_causal_strength : np.ndarray
            Dynamic Causal Strength values
        relative_dynamic_causal_strength : np.ndarray
            Relative Dynamic Causal Strength values
        coefficients : np.ndarray
            VAR coefficients
        """
        super().__init__(
            transfer_entropy=transfer_entropy,
            dynamic_causal_strength=dynamic_causal_strength,
            relative_dynamic_causal_strength=relative_dynamic_causal_strength,
            coefficients=coefficients,
        )


class RelativeDCSCalculator(BaseAnalyzer):
    """
    Relative Dynamic Causal Strength (rDCS) calculator.

    This class implements the Relative Dynamic Causal Strength algorithm
    for quantifying causal relationships relative to a baseline period.
    """

    def __init__(
        self,
        model_order: int,
        reference_time: int,
        estimation_mode: str = "OLS",
        use_diagonal_covariance: bool = False,
        use_old_version: bool = False,
        **kwargs,
    ):
        """
        Initialize Relative DCS calculator.

        Parameters
        ----------
        model_order : int
            Model order for VAR analysis
        reference_time : int
            Reference time point for baseline calculation
        estimation_mode : str
            Estimation mode: 'OLS' or 'RLS'
        use_diagonal_covariance : bool
            Whether to use diagonal covariance approximation
        use_old_version : bool
            Whether to use old version of rDCS calculation
        **kwargs
            Additional configuration parameters
        """
        super().__init__(
            model_order=model_order,
            reference_time=reference_time,
            estimation_mode=estimation_mode,
            use_diagonal_covariance=use_diagonal_covariance,
            use_old_version=use_old_version,
            **kwargs,
        )

    def analyze(  # type: ignore[override]
        self, event_data: np.ndarray, stats: Dict, **kwargs
    ) -> RelativeDCSResult:
        """
        Perform Relative Dynamic Causal Strength analysis.

        Parameters
        ----------
        event_data : np.ndarray
            Event data array of shape (nvar * (model_order + 1), nobs, ntrials)
        stats : Dict
            Model statistics with keys:
            - 'OLS' or 'RLS': Sub-dict with 'At' (coefficients) and 'Sigma_Et' (residual covariance)
            - 'Sigma': Covariance matrices
            - 'mean': Mean values
        **kwargs
            Additional parameters

        Returns
        -------
        RelativeDCSResult
            Relative DCS analysis results
        """
        self._validate_input_data(event_data, stats)
        self._log_analysis_start(event_data.shape)

        try:
            causal_params = {
                "ref_time": self.config["reference_time"],
                "estim_mode": self.config["estimation_mode"],
                "morder": self.config["model_order"],
                "diag_flag": self.config["use_diagonal_covariance"],
                "old_version": self.config["use_old_version"],
            }

            causality_measures = time_varying_causality(
                event_data, stats, causal_params
            )

            self._log_analysis_complete()

            return RelativeDCSResult(
                transfer_entropy=causality_measures["TE"],
                dynamic_causal_strength=causality_measures["DCS"],
                relative_dynamic_causal_strength=causality_measures["rDCS"],
                coefficients=stats[self.config["estimation_mode"]]["At"],
            )

        except Exception as e:
            logger.error(f"Relative DCS analysis failed: {e}")
            raise ComputationError(
                f"Relative DCS analysis failed: {e}",
                "rdcs_computation",
                event_data.shape,
            )

    def _validate_config(self) -> None:
        """Validate configuration parameters."""
        if self.config["model_order"] <= 0:
            raise ValidationError(
                "model_order must be positive",
                "model_order",
                self.config["model_order"],
            )

        if self.config["reference_time"] < 0:
            raise ValidationError(
                "reference_time must be non-negative",
                "reference_time",
                self.config["reference_time"],
            )

        if self.config["estimation_mode"] not in ["OLS", "RLS"]:
            raise ValidationError(
                "estimation_mode must be 'OLS' or 'RLS'",
                "estimation_mode",
                self.config["estimation_mode"],
            )

    def _validate_input_data(  # type: ignore[override]
        self, event_data: np.ndarray, stats: Dict
    ) -> None:
        """Validate input data format and dimensions."""
        super()._validate_input_data(event_data)

        if event_data.ndim != 3:
            raise ValidationError(
                "event_data must be 3D", "event_data_ndim", event_data.ndim
            )

        if not isinstance(stats, dict):
            raise ValidationError(
                "stats must be a dictionary", "stats_type", type(stats)
            )

        required_keys = ["OLS", "Sigma", "mean"]
        for key in required_keys:
            if key not in stats:
                raise ValidationError(
                    f"stats must contain '{key}'", "stats_missing_key", key
                )


def time_varying_causality(
    event_data: np.ndarray, stats: Dict, causal_params: Dict
) -> Dict[str, np.ndarray]:
    """
    Compute time-varying causality measures for bivariate signals.

    Calculates Transfer Entropy (TE), Dynamic Causal Strength (DCS), and Relative Dynamic Causal
    Strength (rDCS) based on a VAR model. This function maintains exact mathematical alignment
    with the previous implementation while providing enhanced error handling and validation.

    Parameters
    ----------
    event_data : np.ndarray
        Event data array of shape (nvar * (model_order + 1), nobs, ntrials).
        Must be bivariate (nvar = 2).
    stats : Dict
        Model statistics with keys:
        - 'OLS' or 'RLS': Sub-dict with 'At' (coefficients) and 'Sigma_Et' (residual covariance).
        - 'Sigma': Covariance matrices of shape
          (nobs, nvar * (model_order + 1), nvar * (model_order + 1)).
        - 'mean': Mean values of shape (nvar * (model_order + 1), nobs).
    causal_params : Dict
        Parameters with keys:
        - 'ref_time': Reference time index for rDCS calculation.
        - 'estim_mode': 'OLS' or 'RLS' estimation mode.
        - 'morder': Model order (number of lags).
        - 'diag_flag': Boolean for diagonal covariance approximation.
        - 'old_version': Boolean for rDCS calculation method.

    Returns
    -------
    Dict[str, np.ndarray]
        Causality measures:
        - 'TE': Transfer Entropy, shape (nobs, 2) where [:, 0] is Y->X, [:, 1] is X->Y.
        - 'DCS': Dynamic Causal Strength, shape (nobs, 2) where [:, 0] is Y->X, [:, 1] is X->Y.
        - 'rDCS': Relative Dynamic Causal Strength, shape (nobs, 2)
          where [:, 0] is Y->X, [:, 1] is X->Y.

    Raises
    ------
    ValidationError
        If input data or parameters are invalid.
    ComputationError
        If computation fails due to numerical issues.

    Notes
    -----
    - The function assumes bivariate data (nvar = 2).
    - Transfer Entropy (TE) measures directed information flow.
    - Dynamic Causal Strength (DCS) measures direct causal influence.
    - Relative Dynamic Causal Strength (rDCS) measures causal influence relative to a baseline.
    - All measures are computed using the Structural Causal Model (SCM) framework.

    Examples
    --------
    >>> event_data = np.random.randn(6, 50, 10)  # (nvar * (morder + 1), nobs, ntrials)
    >>> stats = {
    ...     'OLS': {
    ...         'At': np.random.randn(50, 2, 4),
    ...         'Sigma_Et': np.array([np.eye(2) for _ in range(50)])
    ...     },
    ...     'Sigma': np.random.randn(50, 6, 6),
    ...     'mean': np.random.randn(6, 50)
    ... }
    >>> causal_params = {
    ...     'ref_time': 10, 'estim_mode': 'OLS', 'morder': 2,
    ...     'diag_flag': False, 'old_version': False
    ... }
    >>> result = time_varying_causality(event_data, stats, causal_params)
    >>> print(f"TE shape: {result['TE'].shape}")
    >>> print(f"DCS shape: {result['DCS'].shape}")
    >>> print(f"rDCS shape: {result['rDCS'].shape}")
    """
    _validate_time_varying_causality_inputs(event_data, stats, causal_params)

    _, nobs, ntrials = event_data.shape
    nvar = stats["OLS"]["At"].shape[1]
    ref_time = _normalize_ref_time(causal_params["ref_time"], nobs)
    estim_mode = causal_params["estim_mode"]
    morder = causal_params["morder"]
    diag_flag = causal_params["diag_flag"]
    old_version = causal_params["old_version"]

    logger.info(
        f"Computing time-varying causality: mode={estim_mode}, morder={morder}, ref_time={ref_time}"
    )

    causality_measures = {
        "TE": np.zeros((nobs, 2)),
        "DCS": np.zeros((nobs, 2)),
        "rDCS": np.zeros((nobs, 2)),
    }

    for t in range(nobs):
        try:
            _compute_causality_at_timepoint(
                t,
                event_data,
                stats,
                causality_measures,
                nvar,
                ntrials,
                ref_time,
                estim_mode,
                morder,
                diag_flag,
                old_version,
            )
        except Exception as e:
            logger.error(f"Computation failed at time point {t}: {e}")
            causality_measures["TE"][t] = np.zeros(2)
            causality_measures["DCS"][t] = np.zeros(2)
            causality_measures["rDCS"][t] = np.zeros(2)

    logger.info("Time-varying causality computation completed")
    return causality_measures


def _validate_time_varying_causality_inputs(
    event_data: np.ndarray, stats: Dict, causal_params: Dict
) -> None:
    """
    Validate inputs for time_varying_causality function.

    Parameters
    ----------
    event_data : np.ndarray
        Event data array
    stats : Dict
        Model statistics
    causal_params : Dict
        Causality parameters

    Raises
    ------
    ValidationError
        If inputs are invalid
    """
    if not isinstance(event_data, np.ndarray):
        raise ValidationError(
            "event_data must be a NumPy array", "event_data_type", type(event_data)
        )

    if event_data.ndim != 3:
        raise ValidationError(
            "event_data must be 3D", "event_data_ndim", event_data.ndim
        )

    if not isinstance(stats, dict):
        raise ValidationError("stats must be a dictionary", "stats_type", type(stats))

    required_keys = ["OLS", "Sigma", "mean"]
    for key in required_keys:
        if key not in stats:
            raise ValidationError(
                f"stats must contain '{key}'", "stats_missing_key", key
            )

    if not isinstance(causal_params, dict):
        raise ValidationError(
            "causal_params must be a dictionary",
            "causal_params_type",
            type(causal_params),
        )

    required_params = ["ref_time", "estim_mode", "morder", "diag_flag", "old_version"]
    for param in required_params:
        if param not in causal_params:
            raise ValidationError(
                f"causal_params must contain '{param}'", "causal_params_missing", param
            )

    if causal_params["estim_mode"] not in ["OLS", "RLS"]:
        raise ValidationError(
            "estim_mode must be 'OLS' or 'RLS'",
            "estim_mode",
            causal_params["estim_mode"],
        )

    _validate_ref_time(causal_params["ref_time"])


def _validate_ref_time(ref_time: object) -> None:
    """
    Validate ref_time parameter.

    Parameters
    ----------
    ref_time : object
        Reference time parameter to validate

    Raises
    ------
    ValidationError
        If ref_time is invalid
    """
    if isinstance(ref_time, (int, np.integer)):
        if ref_time < 0:
            raise ValidationError("ref_time must be non-negative", "ref_time", ref_time)
    else:
        try:
            values = list(ref_time)
        except Exception:
            raise ValidationError(
                "ref_time must be an int or an iterable of ints",
                "ref_time",
                type(ref_time),
            )
        if len(values) == 0:
            raise ValidationError(
                "ref_time iterable must be non-empty", "ref_time", values
            )
        if np.any(np.array(values) < 0):
            raise ValidationError(
                "ref_time iterable values must be non-negative", "ref_time", values
            )


def _normalize_ref_time(ref_time: object, nobs: int) -> int:
    """
    Normalize ref_time to an integer suitable for slicing [:ref_time].

    - If ref_time is an int, return it clamped to [0, nobs].
    - If ref_time is an iterable (e.g., range/list/ndarray), use max(ref_time),
      then clamp to [0, nobs].
    """
    if isinstance(ref_time, (int, np.integer)):
        idx = int(ref_time)
    else:
        try:
            values = list(ref_time)
        except Exception:
            raise ValidationError(
                "ref_time must be an int or an iterable of ints",
                "ref_time",
                type(ref_time),
            )
        if len(values) == 0:
            raise ValidationError(
                "ref_time iterable must be non-empty", "ref_time", values
            )
        if np.any(np.array(values) < 0):
            raise ValidationError(
                "ref_time iterable values must be non-negative", "ref_time", values
            )
        idx = int(np.max(values))

    if idx < 0:
        raise ValidationError("ref_time must be non-negative", "ref_time", idx)
    if idx > nobs:
        idx = nobs
    return idx


def _compute_causality_at_timepoint(
    t: int,
    event_data: np.ndarray,
    stats: Dict,
    causality_measures: Dict[str, np.ndarray],
    nvar: int,
    ntrials: int,
    ref_time: int,
    estim_mode: str,
    morder: int,
    diag_flag: bool,
    old_version: bool,
) -> None:
    """
    Compute causality measures for a specific time point.

    Parameters
    ----------
    t : int
        Time point index
    event_data : np.ndarray
        Event data array
    stats : Dict
        Model statistics
    causality_measures : Dict[str, np.ndarray]
        Dictionary to store results
    nvar : int
        Number of variables
    ntrials : int
        Number of trials
    ref_time : int
        Reference time index
    estim_mode : str
        Estimation mode
    morder : int
        Model order
    diag_flag : bool
        Whether to use diagonal covariance
    old_version : bool
        Whether to use old version of rDCS
    """
    lagged_vars = event_data[2:, t, :]

    coeff = stats[estim_mode]["At"][t, :, :]
    residual_cov = stats[estim_mode]["Sigma_Et"][t, :, :]

    a_square = coeff.reshape(nvar, nvar, morder, order="F")
    b = a_square[0, 1, :]  # X -> Y coupling
    c = a_square[1, 0, :]  # Y -> X coupling

    sigy = residual_cov[0, 0] or np.finfo(float).eps
    sigx = residual_cov[1, 1] or np.finfo(float).eps

    x_past_start = 3
    y_past_start = 2
    x_past_indices = slice(x_past_start, x_past_start + 2 * morder, 2)
    y_past_indices = slice(y_past_start, y_past_start + 2 * morder, 2)

    cov_xp = stats["Sigma"][t, x_past_indices, x_past_indices]  # X past covariance
    cov_yp = stats["Sigma"][t, y_past_indices, y_past_indices]  # Y past covariance
    c_xyp = stats["Sigma"][
        t, x_past_indices, y_past_indices
    ]  # X past - Y past cross-covariance
    c_yxp = stats["Sigma"][
        t, y_past_indices, x_past_indices
    ]  # Y past - X past cross-covariance
    mean_xp = stats["mean"][x_past_indices, t]  # X past mean
    mean_yp = stats["mean"][y_past_indices, t]  # Y past mean

    cov_xp_reg = regularize_if_singular(cov_xp)
    cov_yp_reg = regularize_if_singular(cov_yp)

    _compute_transfer_entropy(
        t,
        b,
        c,
        sigy,
        sigx,
        cov_xp,
        cov_yp,
        c_xyp,
        c_yxp,
        cov_xp_reg,
        cov_yp_reg,
        causality_measures,
    )

    mean_x_ref = np.mean(
        stats["mean"][x_past_indices, :ref_time], axis=1
    )  # X past mean
    mean_y_ref = np.mean(
        stats["mean"][y_past_indices, :ref_time], axis=1
    )  # Y past mean

    cov_xp_ref = (
        cov_xp
        + mean_xp @ mean_xp.T
        - mean_xp @ mean_x_ref.T
        - mean_x_ref @ mean_xp.T
        + mean_x_ref @ mean_x_ref.T
    )
    cov_yp_ref = (
        cov_yp
        + mean_yp @ mean_yp.T
        - mean_yp @ mean_y_ref.T
        - mean_y_ref @ mean_yp.T
        + mean_y_ref @ mean_y_ref.T
    )

    ref_cov_xp = np.mean(
        stats["Sigma"][:ref_time, x_past_indices, x_past_indices], axis=0
    )  # X past
    ref_cov_yp = np.mean(
        stats["Sigma"][:ref_time, y_past_indices, y_past_indices], axis=0
    )  # Y past

    _compute_causal_strength_measures(
        t,
        b,
        c,
        sigy,
        sigx,
        cov_xp,
        cov_yp,
        cov_xp_ref,
        cov_yp_ref,
        ref_cov_xp,
        ref_cov_yp,
        lagged_vars,
        stats,
        ntrials,
        ref_time,
        diag_flag,
        old_version,
        causality_measures,
    )


def _compute_transfer_entropy(
    t: int,
    b: np.ndarray,
    c: np.ndarray,
    sigy: float,
    sigx: float,
    cov_xp: np.ndarray,
    cov_yp: np.ndarray,
    c_xyp: np.ndarray,
    c_yxp: np.ndarray,
    cov_xp_reg: np.ndarray,
    cov_yp_reg: np.ndarray,
    causality_measures: Dict[str, np.ndarray],
) -> None:
    """
    Compute Transfer Entropy for a specific time point.

    Parameters
    ----------
    t : int
        Time point index
    b : np.ndarray
        X -> Y coupling coefficients
    c : np.ndarray
        Y -> X coupling coefficients
    sigy : float
        Y residual variance
    sigx : float
        X residual variance
    cov_xp : np.ndarray
        X past covariance
    cov_yp : np.ndarray
        Y past covariance
    c_xyp : np.ndarray
        X past - Y past cross-covariance
    c_yxp : np.ndarray
        Y past - X past cross-covariance
    cov_xp_reg : np.ndarray
        Regularized X past covariance
    cov_yp_reg : np.ndarray
        Regularized Y past covariance
    causality_measures : Dict[str, np.ndarray]
        Dictionary to store results
    """
    # TE(X -> Y)
    causality_measures["TE"][t, 1] = 0.5 * np.log(
        (
            sigy
            + b.T @ cov_xp @ b
            - b.T @ c_xyp @ np.linalg.inv(cov_yp_reg) @ c_xyp.T @ b
        )
        / sigy
    )

    # TE(Y -> X)
    causality_measures["TE"][t, 0] = 0.5 * np.log(
        (
            sigx
            + c.T @ cov_yp @ c
            - c.T @ c_yxp @ np.linalg.inv(cov_xp_reg) @ c_yxp.T @ c
        )
        / sigx
    )


def _compute_causal_strength_measures(
    t: int,
    b: np.ndarray,
    c: np.ndarray,
    sigy: float,
    sigx: float,
    cov_xp: np.ndarray,
    cov_yp: np.ndarray,
    cov_xp_ref: np.ndarray,
    cov_yp_ref: np.ndarray,
    ref_cov_xp: np.ndarray,
    ref_cov_yp: np.ndarray,
    lagged_vars: np.ndarray,
    stats: Dict,
    ntrials: int,
    ref_time: int,
    diag_flag: bool,
    old_version: bool,
    causality_measures: Dict[str, np.ndarray],
) -> None:
    """
    Compute Dynamic Causal Strength (DCS) and Relative Dynamic Causal Strength (rDCS).

    Parameters
    ----------
    t : int
        Time point index
    b : np.ndarray
        X -> Y coupling coefficients
    c : np.ndarray
        Y -> X coupling coefficients
    sigy : float
        Y residual variance
    sigx : float
        X residual variance
    cov_xp : np.ndarray
        X past covariance
    cov_yp : np.ndarray
        Y past covariance
    cov_xp_ref : np.ndarray
        Reference-adjusted X past covariance
    cov_yp_ref : np.ndarray
        Reference-adjusted Y past covariance
    ref_cov_xp : np.ndarray
        Reference X past covariance
    ref_cov_yp : np.ndarray
        Reference Y past covariance
    lagged_vars : np.ndarray
        Lagged variables
    stats : Dict
        Model statistics
    ntrials : int
        Number of trials
    ref_time : int
        Reference time index
    diag_flag : bool
        Whether to use diagonal covariance
    old_version : bool
        Whether to use old version of rDCS
    causality_measures : Dict[str, np.ndarray]
        Dictionary to store results
    """
    if not diag_flag:
        causality_measures["DCS"][t, 1] = 0.5 * np.log((sigy + b.T @ cov_xp @ b) / sigy)
        causality_measures["DCS"][t, 0] = 0.5 * np.log((sigx + c.T @ cov_yp @ c) / sigx)

        if old_version:
            _compute_old_version_rdcs(
                t,
                b,
                c,
                sigy,
                sigx,
                ref_cov_xp,
                ref_cov_yp,
                lagged_vars,
                stats,
                ntrials,
                ref_time,
                causality_measures,
            )
        else:
            _compute_new_version_rdcs(
                t,
                b,
                c,
                sigy,
                sigx,
                cov_xp_ref,
                cov_yp_ref,
                ref_cov_xp,
                ref_cov_yp,
                causality_measures,
            )
    else:
        causality_measures["DCS"][t, 1] = 0.5 * np.log(
            (sigy + b.T @ np.diag(np.diag(cov_xp)) @ b) / sigy
        )
        causality_measures["DCS"][t, 0] = 0.5 * np.log(
            (sigx + c.T @ np.diag(np.diag(cov_yp)) @ c) / sigx
        )

        if old_version:
            _compute_old_version_rdcs_diagonal(
                t,
                b,
                c,
                sigy,
                sigx,
                ref_cov_xp,
                ref_cov_yp,
                lagged_vars,
                stats,
                ntrials,
                ref_time,
                causality_measures,
            )
        else:
            _compute_new_version_rdcs_diagonal(
                t,
                b,
                c,
                sigy,
                sigx,
                cov_xp_ref,
                cov_yp_ref,
                ref_cov_xp,
                ref_cov_yp,
                causality_measures,
            )


def _compute_old_version_rdcs(
    t: int,
    b: np.ndarray,
    c: np.ndarray,
    sigy: float,
    sigx: float,
    ref_cov_xp: np.ndarray,
    ref_cov_yp: np.ndarray,
    lagged_vars: np.ndarray,
    stats: Dict,
    ntrials: int,
    ref_time: int,
    causality_measures: Dict[str, np.ndarray],
) -> None:
    """Compute old version of rDCS with full covariance."""
    cov_xp_lag = (
        np.dot(
            lagged_vars - np.mean(stats["mean"][3:, :ref_time], axis=1),
            (lagged_vars - np.mean(stats["mean"][3:, :ref_time], axis=1)).T,
        )
        / ntrials
    )

    causality_measures["rDCS"][t, 1] = (
        0.5 * np.log((sigy + b.T @ ref_cov_xp @ b) / sigy)
        - 0.5
        + 0.5
        * (sigy + b.T @ cov_xp_lag[2::2, 2::2] @ b)
        / (sigy + b.T @ ref_cov_xp @ b)
    )

    causality_measures["rDCS"][t, 0] = (
        (0.5 * np.log((sigx + c.T @ ref_cov_yp @ c) / sigx))
        - 0.5
        + (
            0.5
            * (sigx + c.T @ cov_xp_lag[1::2, 1::2] @ c)
            / (sigx + c.T @ ref_cov_yp @ c)
        )
    )


def _compute_new_version_rdcs(
    t: int,
    b: np.ndarray,
    c: np.ndarray,
    sigy: float,
    sigx: float,
    cov_xp_ref: np.ndarray,
    cov_yp_ref: np.ndarray,
    ref_cov_xp: np.ndarray,
    ref_cov_yp: np.ndarray,
    causality_measures: Dict[str, np.ndarray],
) -> None:
    """Compute new version of rDCS with full covariance."""
    causality_measures["rDCS"][t, 1] = (
        (0.5 * np.log((sigy + b.T @ ref_cov_xp @ b) / sigy))
        - 0.5
        + (0.5 * (sigy + b.T @ cov_xp_ref @ b) / (sigy + b.T @ ref_cov_xp @ b))
    )

    causality_measures["rDCS"][t, 0] = (
        (0.5 * np.log((sigx + c.T @ ref_cov_yp @ c) / sigx))
        - 0.5
        + (0.5 * (sigx + c.T @ cov_yp_ref @ c) / (sigx + c.T @ ref_cov_yp @ c))
    )


def _compute_old_version_rdcs_diagonal(
    t: int,
    b: np.ndarray,
    c: np.ndarray,
    sigy: float,
    sigx: float,
    ref_cov_xp: np.ndarray,
    ref_cov_yp: np.ndarray,
    lagged_vars: np.ndarray,
    stats: Dict,
    ntrials: int,
    ref_time: int,
    causality_measures: Dict[str, np.ndarray],
) -> None:
    """Compute old version of rDCS with diagonal covariance."""
    cov_xp_lag = (
        np.dot(
            lagged_vars - np.mean(stats["mean"][3:, :ref_time], axis=1)[:, np.newaxis],
            (
                lagged_vars
                - np.mean(stats["mean"][3:, :ref_time], axis=1)[:, np.newaxis]
            ).T,
        )
        / ntrials
    )

    causality_measures["rDCS"][t, 1] = (
        (0.5 * np.log((sigy + b.T @ np.diag(np.diag(ref_cov_xp)) @ b) / sigy))
        - 0.5
        + (
            0.5
            * (sigy + b.T @ np.diag(np.diag(cov_xp_lag[2::2, 2::2])) @ b)
            / (sigy + b.T @ np.diag(np.diag(ref_cov_xp)) @ b)
        )
    )

    causality_measures["rDCS"][t, 0] = (
        (0.5 * np.log((sigx + c.T @ np.diag(np.diag(ref_cov_yp)) @ c) / sigx))
        - 0.5
        + (
            0.5
            * (sigx + c.T @ np.diag(np.diag(cov_xp_lag[1::2, 1::2])) @ c)
            / (sigx + c.T @ np.diag(np.diag(ref_cov_yp)) @ c)
        )
    )


def _compute_new_version_rdcs_diagonal(
    t: int,
    b: np.ndarray,
    c: np.ndarray,
    sigy: float,
    sigx: float,
    cov_xp_ref: np.ndarray,
    cov_yp_ref: np.ndarray,
    ref_cov_xp: np.ndarray,
    ref_cov_yp: np.ndarray,
    causality_measures: Dict[str, np.ndarray],
) -> None:
    """Compute new version of rDCS with diagonal covariance."""
    causality_measures["rDCS"][t, 1] = (
        (0.5 * np.log((sigy + b.T @ np.diag(np.diag(ref_cov_xp)) @ b) / sigy))
        - 0.5
        + (
            0.5
            * (sigy + b.T @ np.diag(np.diag(cov_xp_ref)) @ b)
            / (sigy + b.T @ np.diag(np.diag(ref_cov_xp)) @ b)
        )
    )

    causality_measures["rDCS"][t, 0] = (
        (0.5 * np.log((sigx + c.T @ np.diag(np.diag(ref_cov_yp)) @ c) / sigx))
        - 0.5
        + (
            0.5
            * (sigx + c.T @ np.diag(np.diag(cov_yp_ref)) @ c)
            / (sigx + c.T @ np.diag(np.diag(ref_cov_yp)) @ c)
        )
    )



"""
Configuration classes for the Dynamic Causal Strength (DCS) package.

This module provides dataclasses for configuring the analysis pipeline,
including detection parameters, BIC model selection, causality analysis,
and output settings.
"""

from dataclasses import dataclass, field
from typing import Dict, List, Optional

import numpy as np


@dataclass
class PipelineOptions:
    """
    Options to control which pipeline steps are executed.

    Attributes
    ----------
    detection : bool
        Whether to perform event detection (default: True)
    bic : bool
        Whether to perform BIC model selection (default: False)
    causal_analysis : bool
        Whether to perform causality analysis (default: True)
    bootstrap : bool
        Whether to perform bootstrap analysis (default: False)
    save_flag : bool
        Whether to save results to files (default: False)
    debiased_stats : bool
        Whether to perform DeSnap analysis (default: True)
    """

    detection: bool = True
    bic: bool = False
    causal_analysis: bool = True
    bootstrap: bool = False
    save_flag: bool = False
    debiased_stats: bool = True


@dataclass
class DetectionParams:
    """
    Parameters for the event detection step.

    Attributes
    ----------
    thres_ratio : float
        Threshold ratio for event detection
    align_type : str
        Alignment type: 'peak' or 'pooled'
    l_extract : int
        Length of extracted snapshots
    l_start : int
        Start offset for snapshot extraction
    shrink_flag : bool
        Whether to use shrinking for pooled alignment (default: False)
    locs : Optional[np.ndarray]
        Pre-provided event locations (default: None)
    remove_artif : bool
        Whether to remove artifact trials (default: False)
    """

    thres_ratio: float
    align_type: str  # 'peak' or 'pooled'
    l_extract: int
    l_start: int
    shrink_flag: bool = False
    locs: Optional[np.ndarray] = None
    remove_artif: bool = False
    remove_artif_threshold: float = -15000


@dataclass
class BicParams:
    """
    Parameters for BIC model order selection.

    Attributes
    ----------
    morder : int
        Model order to use if BIC is False, or default
    momax : Optional[int]
        Max order to test if BIC is True (default: None)
    tau : Optional[int]
        Lag step if BIC is True (default: None)
    mode : Optional[str]
        BIC mode, e.g., 'biased' (default: None)
    """

    morder: int
    momax: Optional[int] = None
    tau: Optional[int] = None
    mode: Optional[str] = None
    estim_mode: Optional[str] = None


@dataclass
class CausalParams:
    """
    Parameters for causality calculation.

    Attributes
    ----------
    ref_time : int
        Reference time point for causality analysis
    estim_mode : str
        Estimation mode: 'OLS' or 'RLS' (default: 'OLS')
    diag_flag : bool
        Whether to use diagonal covariance (default: False)
    old_version : bool
        Whether to use old version of rDCS calculation (default: False)
    """

    ref_time: int
    estim_mode: str = "OLS"  # 'OLS' or 'RLS'
    diag_flag: bool = False
    old_version: bool = False


@dataclass
class MonteCParams:
    """
    Parameters for Monte Carlo bootstrapping.

    Attributes
    ----------
    n_btsp : int
        Number of bootstrap samples (default: 100)
    """

    n_btsp: int = 100


@dataclass
class OutputParams:
    """
    Parameters for output file naming.

    Attributes
    ----------
    file_keyword : str
        Keyword for output file naming
    save_path : str
        Path for saving output files
    """

    file_keyword: str
    save_path: str = ""


@dataclass
class DeSnapParams:
    """
    Input structure for the de-snapshotting analysis.

    Attributes
    ----------
    detection_signal : np.ndarray
        Conditioning variable values (1D array)
    original_signal : np.ndarray
        Original time series data
    Yt_stats_cond : Dict
        Conditional statistics from pipeline
    morder : int
        Model order
    tau : int
        Lag step
    l_start : int
        Start offset for snapshot extraction
    l_extract : int
        Length of extracted snapshots
    d0 : float
        Lower bound of the first bin for D
    N_d : int
        Number of bins for D
    d0_max : Optional[float]
        Upper bound for binning D (default: None)
    maxStdRatio : Optional[float]
        Alternative to define d0_max based on std(D) (default: None)
    diff_flag : bool
        Flag for how 'c' (covariance adjustment factor) is calculated (default: False)
    """

    detection_signal: np.ndarray
    original_signal: np.ndarray
    event_stats: Dict
    morder: int
    tau: int
    l_start: int
    l_extract: int
    d0: float
    N_d: int
    d0_max: Optional[float] = None
    maxStdRatio: Optional[float] = None
    diff_flag: bool = False


@dataclass
class PipelineConfig:
    """
    Main configuration object for the analysis pipeline.

    This class combines all configuration parameters into a single object
    for easy management and validation of pipeline settings.

    Attributes
    ----------
    options : PipelineOptions
        Pipeline execution options
    detection : DetectionParams
        Event detection parameters
    bic : BicParams
        BIC model selection parameters
    causal : CausalParams
        Causality analysis parameters
    output : OutputParams
        Output file parameters
    monte_carlo : Optional[MonteCParams]
        Monte Carlo bootstrap parameters (default: None)
    desnap : Optional[DeSnapParams]
        DeSnap analysis parameters (default: None)
    sampling_rate : int
        Sampling rate in Hz (default: 1252)
    passband : List[int]
        Frequency passband [low, high] in Hz (default: [140, 230])
    """

    options: PipelineOptions
    detection: DetectionParams
    bic: BicParams
    causal: CausalParams
    output: OutputParams
    monte_carlo: Optional[MonteCParams] = None
    desnap: Optional[DeSnapParams] = None
    sampling_rate: int = 1252
    passband: List[int] = field(default_factory=lambda: [140, 230])

    def __post_init__(self) -> None:
        """Validate configuration parameters after initialization."""
        self._validate_detection_params()
        self._validate_bic_params()
        self._validate_bootstrap_params()
        self._validate_align_type()

    def _validate_detection_params(self) -> None:
        """Validate detection parameters."""
        if not self.options.detection and self.detection.locs is None:
            raise ValueError(
                "detection.locs must be provided if options.detection is False"
            )

    def _validate_bic_params(self) -> None:
        """Validate BIC parameters."""
        if self.options.bic and (
            self.bic.momax is None or self.bic.tau is None or self.bic.mode is None
        ):
            raise ValueError(
                "bic.momax, bic.tau, and bic.mode must be set if options.bic is True"
            )

    def _validate_bootstrap_params(self) -> None:
        """Validate bootstrap parameters."""
        if self.options.bootstrap and self.monte_carlo is None:
            raise ValueError(
                "monte_carlo parameters must be provided if options.bootstrap is True"
            )

    def _validate_align_type(self) -> None:
        """Validate alignment type."""
        if self.detection.align_type not in ["peak", "pooled"]:
            raise ValueError(
                f"Invalid detection.align_type: {self.detection.align_type}. "
                "Must be 'peak' or 'pooled'."
            )



"""
Model selection utilities for Dynamic Causal Strength (DCS).

This module provides high-level model selection utilities including
BIC-based model order selection and related functionality.
"""

import logging
from typing import Tuple

import numpy as np

from .var_estimation import VAREstimator

logger = logging.getLogger(__name__)


def select_model_order(
    time_series_data: np.ndarray, max_model_order: int, time_mode: str
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Select the optimal VAR model order using Bayesian Information Criterion (BIC).

    Evaluates BIC scores for model orders 1 to max_model_order under specified time mode.

    Parameters
    ----------
    time_series_data : np.ndarray
        Time series data, shape (n_vars, n_observations, n_trials).
    max_model_order : int
        Maximum model order to evaluate.
    time_mode : str
        'inhomo' (inhomogeneous) or 'homo' (homogeneous).

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        - bic_scores: BIC scores, shape (max_model_order, 2).
        - optimal_orders: Optimal model orders for each BIC variant, shape (2,).
        - log_likelihoods: Log-likelihoods, shape (max_model_order,).
        - penalty_terms: Penalty terms, shape (max_model_order, 2).

    Raises
    ------
    ValueError
        If time_mode is invalid or all BIC scores are NaN.

    Examples
    --------
    >>> import numpy as np
    >>> data = np.random.randn(2, 1000, 20)  # (n_vars, n_obs, n_trials)
    >>> bic_scores, optimal_orders, log_likelihoods, penalty_terms = select_model_order(
    ...     data, max_model_order=6, time_mode="inhomo"
    ... )
    >>> print(f"Optimal model orders: {optimal_orders}")
    """
    if time_mode not in ["inhomo", "homo"]:
        raise ValueError("time_mode must be 'inhomo' or 'homo'.")

    n_vars, n_observations, n_trials = time_series_data.shape
    bic_scores = np.full((max_model_order, 2), np.nan)
    penalty_terms = np.full((max_model_order, 2), np.nan)
    log_likelihoods = np.full(max_model_order, np.nan)
    sum_log_det_hessian = np.full(max_model_order, np.nan)

    for model_order in range(1, max_model_order + 1):
        n_time_steps = (
            n_observations - model_order - 1
        )  # Adjusted based on data usage: T = len(range(mo + 1, nobs))
        logger.info(f"Processing model order: {model_order}")

        # Create VAREstimator instance for this model order
        estimator = VAREstimator(model_order=model_order, time_mode=time_mode)

        _, _, log_likelihoods[model_order - 1], sum_log_det_hessian[model_order - 1] = (
            estimator.estimate_var_coefficients(
                time_series_data, model_order, max_model_order, time_mode, "infocrit"
            )
        )

        penalty_terms[model_order - 1, 1] = sum_log_det_hessian[model_order - 1]

        if time_mode == "inhomo":
            penalty_terms[model_order - 1, 0] = (
                n_time_steps * model_order * n_vars * n_vars * np.log(n_trials)
            )
            bic_scores[model_order - 1, 0] = (
                -log_likelihoods[model_order - 1] * n_trials
                + penalty_terms[model_order - 1, 0]
            )
            bic_scores[model_order - 1, 1] = (
                -log_likelihoods[model_order - 1] * n_trials
                + sum_log_det_hessian[model_order - 1]
            )
        elif time_mode == "homo":
            penalty_terms[model_order - 1, 0] = (
                model_order * n_vars * n_vars * np.log(n_time_steps * n_trials)
            )
            bic_scores[model_order - 1, 0] = (
                -log_likelihoods[model_order - 1] * n_trials
                + penalty_terms[model_order - 1, 0]
            )
            bic_scores[model_order - 1, 1] = (
                -log_likelihoods[model_order - 1] * n_trials
                + sum_log_det_hessian[model_order - 1]
            )

    if np.isnan(bic_scores).all():
        raise ValueError(
            "All BIC scores are NaN; verify input data or reduce max_model_order."
        )

    optimal_orders = np.nanargmin(bic_scores, axis=0) + 1
    logger.info(f"Optimal model orders: {optimal_orders}")

    return bic_scores, optimal_orders, log_likelihoods, penalty_terms



"""
TranCIT: Transient Causal Interaction Toolbox

A Python package for quantifying causal relationships in multivariate time series data
using dynamic causal strength (DCS) methods.

This package provides methods for analyzing directional influences using model-based
statistical tools, inspired by information-theoretic and autoregressive frameworks.

The package implements several causality measures:
- Dynamic Causal Strength (DCS): Time-varying causal relationships
- Transfer Entropy (TE): Information-theoretic causality measures
- Granger Causality (GC): Linear causality detection
- Relative Dynamic Causal Strength (rDCS): Event-based causality

Example
-------
>>> import numpy as np
>>> from trancit import DCSCalculator
>>>
>>> # Create calculator
>>> calculator = DCSCalculator(model_order=4, time_mode="inhomo")
>>>
>>> # Generate sample data
>>> data = np.random.randn(2, 1000, 20)  # (n_vars, n_obs, n_trials)
>>>
>>> # Perform analysis
>>> result = calculator.analyze(data)
>>> print(f"DCS shape: {result.causal_strength.shape}")
"""

# Version is automatically set by setuptools_scm from git tags
try:
    from ._version import version as __version__
except ImportError:
    # Fallback for development installs without setuptools_scm
    __version__ = "0.1.0-dev"
__author__ = "Salar Nouri"
__email__ = "salr.nouri@gmail.com"
__license__ = "BSD-2-Clause"

# Causality analysis imports
from trancit.causality.dcs import DCSCalculator, DCSResult
from trancit.causality.granger import GrangerCausalityCalculator, GrangerCausalityResult
from trancit.causality.rdcs import (
    RelativeDCSCalculator,
    RelativeDCSResult,
    time_varying_causality,
)
from trancit.causality.transfer_entropy import (
    TransferEntropyCalculator,
    TransferEntropyResult,
)

# Configuration imports
from trancit.config import (
    BicParams,
    CausalParams,
    DeSnapParams,
    DetectionParams,
    MonteCParams,
    OutputParams,
    PipelineConfig,
    PipelineOptions,
)

# Core functionality imports
from trancit.core.base import BaseAnalyzer, BaseConfig, BaseResult
from trancit.core.exceptions import (
    ComputationError,
    ConfigurationError,
    ConvergenceError,
    DataError,
    DCSError,
    SingularMatrixError,
    ValidationError,
)

# Setup logging
from trancit.logger_config import setup_logging
from trancit.models.bic_selection import BICSelector
from trancit.models.model_validation import ModelValidator

# Model estimation imports
from trancit.models.var_estimation import VAREstimator

# Pipeline imports
from trancit.pipeline.orchestrator import PipelineOrchestrator, PipelineResult
from trancit.pipeline.stages import (
    ArtifactRemovalStage,
    BICSelectionStage,
    BootstrapAnalysisStage,
    BorderRemovalStage,
    CausalityAnalysisStage,
    DeSnapAnalysisStage,
    EventDetectionStage,
    InputValidationStage,
    OutputPreparationStage,
    SnapshotExtractionStage,
    StatisticsComputationStage,
)

# Simulation imports
from trancit.simulation import (
    generate_ensemble_nonstat_innomean,
    generate_signals,
    generate_var_nonstat,
    morlet,
    simulate_ar_event,
    simulate_ar_event_bootstrap,
    simulate_ar_nonstat_innomean,
)

# Utility imports
from trancit.utils import (
    compute_event_statistics,
    estimate_residuals,
    extract_event_snapshots,
    find_best_shrinked_locations,
    find_peak_locations,
    get_residuals,
    perform_desnap_analysis,
    remove_artifact_trials,
    shrink_locations_resample_uniform,
)

setup_logging(log_file="trancit_log.txt")

# Public API definition
__all__ = [
    # Core classes
    "BaseAnalyzer",
    "BaseResult",
    "BaseConfig",
    # Exceptions
    "DCSError",
    "ValidationError",
    "ComputationError",
    "ConfigurationError",
    "DataError",
    "ConvergenceError",
    "SingularMatrixError",
    # Configuration classes
    "PipelineConfig",
    "PipelineOptions",
    "DetectionParams",
    "BicParams",
    "CausalParams",
    "MonteCParams",
    "OutputParams",
    "DeSnapParams",
    # Causality analysis
    "DCSCalculator",
    "DCSResult",
    "TransferEntropyCalculator",
    "TransferEntropyResult",
    "GrangerCausalityCalculator",
    "GrangerCausalityResult",
    "RelativeDCSCalculator",
    "RelativeDCSResult",
    "time_varying_causality",
    # Model estimation
    "VAREstimator",
    "BICSelector",
    "ModelValidator",
    # Pipeline stages
    "PipelineOrchestrator",
    "PipelineResult",
    "InputValidationStage",
    "EventDetectionStage",
    "BorderRemovalStage",
    "BICSelectionStage",
    "SnapshotExtractionStage",
    "ArtifactRemovalStage",
    "StatisticsComputationStage",
    "CausalityAnalysisStage",
    "BootstrapAnalysisStage",
    "DeSnapAnalysisStage",
    "OutputPreparationStage",
    # Simulation functions
    "generate_signals",
    "simulate_ar_event",
    "simulate_ar_event_bootstrap",
    "simulate_ar_nonstat_innomean",
    "generate_ensemble_nonstat_innomean",
    "generate_var_nonstat",
    "morlet",
    # Utility functions
    "compute_event_statistics",
    "perform_desnap_analysis",
    "estimate_residuals",
    "extract_event_snapshots",
    "get_residuals",
    "remove_artifact_trials",
    "find_peak_locations",
    "shrink_locations_resample_uniform",
    "find_best_shrinked_locations",
]



"""
Transfer Entropy (TE) implementation.

This module provides the implementation of Transfer Entropy
calculation using the new modular architecture.
"""

import logging
from typing import Dict, Tuple

import numpy as np

from trancit.core.base import BaseAnalyzer, BaseResult
from trancit.core.exceptions import ComputationError, ValidationError
from trancit.utils.helpers import compute_covariances, estimate_coefficients
from trancit.utils.preprocess import regularize_if_singular

logger = logging.getLogger(__name__)


class TransferEntropyResult(BaseResult):
    """Result container for Transfer Entropy analysis."""

    def __init__(
        self,
        transfer_entropy: np.ndarray,
        te_residual_cov: np.ndarray,
        coefficients: np.ndarray,
    ):
        """
        Initialize Transfer Entropy result.

        Parameters
        ----------
        transfer_entropy : np.ndarray
            Transfer Entropy values
        te_residual_cov : np.ndarray
            Transfer entropy residual covariance
        coefficients : np.ndarray
            VAR coefficients
        """
        super().__init__(
            transfer_entropy=transfer_entropy,
            te_residual_cov=te_residual_cov,
            coefficients=coefficients,
        )


class TransferEntropyCalculator(BaseAnalyzer):
    """
    Transfer Entropy (TE) calculator.

    This class implements the Transfer Entropy algorithm for quantifying
    information flow between time series variables.
    """

    def __init__(
        self,
        model_order: int,
        time_mode: str = "inhomo",
        use_diagonal_covariance: bool = False,
        **kwargs,
    ):
        """
        Initialize Transfer Entropy calculator.

        Parameters
        ----------
        model_order : int
            Model order for VAR analysis
        time_mode : str
            Time mode: 'inhomo' (time-inhomogeneous) or 'homo' (homogeneous)
        use_diagonal_covariance : bool
            Whether to use diagonal covariance approximation
        **kwargs
            Additional configuration parameters
        """
        super().__init__(
            model_order=model_order,
            time_mode=time_mode,
            use_diagonal_covariance=use_diagonal_covariance,
            **kwargs,
        )
        # VAREstimator is not used in this implementation - using direct computation

    def analyze(self, data: np.ndarray, **kwargs) -> TransferEntropyResult:
        """
        Perform Transfer Entropy analysis on the given data.

        Parameters
        ----------
        data : np.ndarray
            Time series data with shape (n_vars, n_observations, n_trials)
        **kwargs
            Additional parameters

        Returns
        -------
        TransferEntropyResult
            Transfer Entropy analysis results
        """
        self._validate_input_data(data)
        self._log_analysis_start(data.shape)

        try:
            current_data, lagged_data, n_time_steps = self._prepare_data_structures(
                data
            )

            cov_xp, cov_yp, c_xyp, c_yxp = compute_covariances(
                lagged_data, n_time_steps, self.config["model_order"]
            )

            result_arrays = self._initialize_result_arrays(
                n_time_steps, data.shape[1], data.shape[0]
            )

            if self.config["time_mode"] == "inhomo":
                self._compute_inhomogeneous_te(
                    current_data,
                    lagged_data,
                    cov_xp,
                    cov_yp,
                    c_xyp,
                    c_yxp,
                    result_arrays,
                    data.shape[2],
                    data.shape[0],
                )
                self._adjust_outputs_for_inhomo(result_arrays)
            else:
                self._compute_homogeneous_te(
                    current_data,
                    lagged_data,
                    cov_xp,
                    cov_yp,
                    c_xyp,
                    c_yxp,
                    result_arrays,
                    data.shape[2],
                    data.shape[0],
                )

            self._log_analysis_complete()

            return TransferEntropyResult(
                transfer_entropy=result_arrays["transfer_entropy"],
                te_residual_cov=result_arrays["te_residual_cov"],
                coefficients=result_arrays["coefficients"],
            )

        except Exception as e:
            logger.error(f"Transfer Entropy analysis failed: {e}")
            raise ComputationError(
                f"Transfer Entropy analysis failed: {e}", "te_computation", data.shape
            )

    def _validate_config(self) -> None:
        """Validate configuration parameters."""
        if self.config["model_order"] <= 0:
            raise ValidationError(
                "model_order must be positive",
                "model_order",
                self.config["model_order"],
            )

        if self.config["time_mode"] not in ["inhomo", "homo"]:
            raise ValidationError(
                "time_mode must be 'inhomo' or 'homo'",
                "time_mode",
                self.config["time_mode"],
            )

    def _validate_input_data(self, data: np.ndarray) -> None:
        """Validate input data format and dimensions."""
        super()._validate_input_data(data)

        if data.ndim != 3:
            raise ValidationError(
                "Input data must be 3D (n_vars, n_observations, n_trials)",
                "data_ndim",
                data.ndim,
            )

        if data.shape[0] != 2:
            raise ValidationError(
                "Input data must be bivariate (n_vars=2)", "n_vars", data.shape[0]
            )

    def _prepare_data_structures(
        self, time_series_data: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, int]:
        """
        Prepare data structures for analysis.

        Parameters
        ----------
        time_series_data : np.ndarray
            Input time series data

        Returns
        -------
        Tuple[np.ndarray, np.ndarray, int]
            current_data, lagged_data, n_time_steps
        """
        nvar, nobs, ntrials = time_series_data.shape
        r1 = self.config["model_order"] + 1

        extended_data = np.zeros((nvar, r1, nobs + r1 - 1, ntrials))
        for k in range(r1):
            extended_data[:, k, k : k + nobs, :] = time_series_data

        current_data = extended_data[:, 0, r1 - 1 : nobs, :]
        lagged_data = extended_data[
            :, 1 : self.config["model_order"] + 1, r1 - 1 : nobs - 1, :
        ]
        n_time_steps = current_data.shape[1] - 1

        return current_data, lagged_data, n_time_steps

    def _initialize_result_arrays(
        self, n_time_steps: int, nobs: int, nvar: int
    ) -> Dict[str, np.ndarray]:
        """
        Initialize arrays for storing results.

        Parameters
        ----------
        n_time_steps : int
            Number of time steps
        nobs : int
            Number of observations
        nvar : int
            Number of variables

        Returns
        -------
        Dict[str, np.ndarray]
            Dictionary of initialized arrays
        """
        return {
            "te_residual_cov": np.zeros((n_time_steps, 2)),
            "transfer_entropy": np.zeros((n_time_steps, 2)),
            "coefficients": np.full(
                (nobs, nvar, nvar * self.config["model_order"] + 1), np.nan
            ),
        }

    def _compute_inhomogeneous_te(
        self,
        current_data: np.ndarray,
        lagged_data: np.ndarray,
        cov_xp: np.ndarray,
        cov_yp: np.ndarray,
        c_xyp: np.ndarray,
        c_yxp: np.ndarray,
        result_arrays: Dict[str, np.ndarray],
        ntrials: int,
        nvar: int,
    ) -> None:
        """Compute inhomogeneous Transfer Entropy."""
        n_time_steps = current_data.shape[1] - 1

        for t in range(n_time_steps):
            try:
                coeff, residual_cov = estimate_coefficients(
                    current_data[:, t, :].T,
                    lagged_data[:, : self.config["model_order"], t, :]
                    .reshape(nvar * self.config["model_order"], ntrials)
                    .T,
                    ntrials,
                )

                result_arrays["coefficients"][t, :, :] = coeff

                a_square = coeff[:, :-1].reshape(nvar, nvar, self.config["model_order"])
                b = a_square[0, 1, :]  # X -> Y
                c = a_square[1, 0, :]  # Y -> X
                sigy = residual_cov[0, 0] + np.finfo(float).eps
                sigx = residual_cov[1, 1] + np.finfo(float).eps

                cov_yp_reg = regularize_if_singular(cov_yp[t])
                cov_xp_reg = regularize_if_singular(cov_xp[t])

                result_arrays["te_residual_cov"][t, 1] = (
                    sigy
                    + b.T @ cov_xp_reg @ b
                    - b.T @ c_xyp[t] @ np.linalg.inv(cov_yp_reg) @ c_xyp[t].T @ b
                )
                result_arrays["te_residual_cov"][t, 0] = (
                    sigx
                    + c.T @ cov_yp_reg @ c
                    - c.T @ c_yxp[t] @ np.linalg.inv(cov_xp_reg) @ c_yxp[t].T @ c
                )

                result_arrays["transfer_entropy"][t, 1] = 0.5 * np.log(
                    result_arrays["te_residual_cov"][t, 1] / sigy
                )
                result_arrays["transfer_entropy"][t, 0] = 0.5 * np.log(
                    result_arrays["te_residual_cov"][t, 0] / sigx
                )

            except Exception as e:
                logger.error(
                    f"Transfer Entropy computation failed at time step {t}: {e}"
                )
                result_arrays["transfer_entropy"][t] = np.zeros(2)
                result_arrays["te_residual_cov"][t] = np.zeros(2)

    def _compute_homogeneous_te(
        self,
        current_data: np.ndarray,
        lagged_data: np.ndarray,
        cov_xp: np.ndarray,
        cov_yp: np.ndarray,
        c_xyp: np.ndarray,
        c_yxp: np.ndarray,
        result_arrays: Dict[str, np.ndarray],
        ntrials: int,
        nvar: int,
    ) -> None:
        """Compute homogeneous Transfer Entropy."""
        coeff, residual_cov = estimate_coefficients(current_data, lagged_data, ntrials)

        for t in range(result_arrays["coefficients"].shape[0]):
            result_arrays["coefficients"][t] = coeff.T

        a_square = coeff[:, :-1].reshape(nvar, nvar, self.config["model_order"])
        b = a_square[0, 1, :]  # X -> Y
        c = a_square[1, 0, :]  # Y -> X
        sigy = residual_cov[0, 0] + np.finfo(float).eps
        sigx = residual_cov[1, 1] + np.finfo(float).eps

        cov_xp_avg = np.mean(cov_xp, axis=0)
        cov_yp_avg = np.mean(cov_yp, axis=0)
        cov_xy_p_avg = np.mean(c_xyp, axis=0)
        cov_yx_p_avg = np.mean(c_yxp, axis=0)

        result_arrays["transfer_entropy"][0, 1] = 0.5 * np.log(
            (
                sigy
                + b.T @ cov_xp_avg @ b
                - b.T @ cov_xy_p_avg @ np.linalg.inv(cov_yp_avg) @ cov_xy_p_avg.T @ b
            )
            / sigy
        )
        result_arrays["transfer_entropy"][0, 0] = 0.5 * np.log(
            (
                sigx
                + c.T @ cov_yp_avg @ c
                - c.T @ cov_yx_p_avg @ np.linalg.inv(cov_xp_avg) @ cov_yx_p_avg.T @ c
            )
            / sigx
        )

    def _adjust_outputs_for_inhomo(self, result_arrays: Dict[str, np.ndarray]) -> None:
        """Adjust outputs for inhomogeneous mode by adding NaN blocks."""
        if self.config["time_mode"] == "inhomo":
            nan_block = np.full((self.config["model_order"], 2), np.nan)
            result_arrays["transfer_entropy"] = np.vstack(
                [nan_block, result_arrays["transfer_entropy"]]
            )
            result_arrays["te_residual_cov"] = np.vstack(
                [nan_block, result_arrays["te_residual_cov"]]
            )



"""
Simulation module for the TranCIT package.

This module provides various simulation functions for generating synthetic time series data
for testing and demonstration purposes.

Submodules:
- ar_simulation: Autoregressive process simulations
- oscillator_simulation: Coupled oscillator simulations
- var_simulation: Vector autoregressive simulations
- utils: Simulation utilities (wavelets, etc.)
"""

from .ar_simulation import (
    simulate_ar_event,
    simulate_ar_event_bootstrap,
    simulate_ar_nonstat_innomean,
)
from .oscillator_simulation import generate_signals
from .utils import morlet
from .var_simulation import (
    generate_ensemble_nonstat_innomean,
    generate_var_nonstat,
)

__all__ = [
    "generate_signals",
    "simulate_ar_event",
    "simulate_ar_event_bootstrap",
    "simulate_ar_nonstat_innomean",
    "generate_ensemble_nonstat_innomean",
    "generate_var_nonstat",
    "morlet",
]



import logging
import sys
from typing import Optional

LOG_LEVEL_DEFAULT = logging.INFO
LOG_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
LOGGER_NAME = "trancit"


def setup_logging(
    name: str = LOGGER_NAME,
    level: int = LOG_LEVEL_DEFAULT,
    log_format: str = LOG_FORMAT,
    date_format: str = DATE_FORMAT,
    log_file: Optional[str] = None,
) -> logging.Logger:
    """
    Configures logging for the 'trancit' package.

    This should be called once at application start or from __init__.py.

    Args:
        name (str): Logger name (usually use __name__ or a fixed package name).
        log_file (str): File to write logs to.
        level (int): Logging level.
        log_format (str): Log message format.
        date_format (str): Timestamp format.
    """
    logger = logging.getLogger(name)

    if logger.hasHandlers():
        return logger

    logger.setLevel(level)

    formatter = logging.Formatter(log_format, datefmt=date_format)

    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(level)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    if log_file:
        file_handler = logging.FileHandler(log_file, mode="a")
        file_handler.setFormatter(formatter)
        file_handler.setLevel(level)
        logger.addHandler(file_handler)

    logger.info("Logging configured for '%s'", name)
    return logger



"""
Core utilities for Dynamic Causal Strength (DCS).

This module provides fundamental utilities for event extraction, statistics computation,
and DeSnap analysis. These functions form the backbone of the DCS pipeline.
"""

import logging
from typing import Dict, Union

import numpy as np

from ..config import DeSnapParams
from ..core.exceptions import ComputationError, ValidationError
from .preprocess import regularize_if_singular
from .residuals import estimate_residuals

logger = logging.getLogger(__name__)


def extract_event_windows(
    signal: np.ndarray, centers: np.ndarray, start_offset: int, window_length: int
) -> np.ndarray:
    """
    Extract windows of data from a signal around specified center points.

    This function extracts fixed-length windows from a 1D signal around each
    specified center point. The windows are extracted with a specified offset
    from the center and have a fixed length.

    Parameters
    ----------
    signal : np.ndarray
        1D array representing the signal data.
    centers : np.ndarray
        1D array of center points (indices) around which to extract windows.
    start_offset : int
        Offset from the center to start the window (can be negative).
    window_length : int
        Length of each window to extract.

    Returns
    -------
    np.ndarray
        2D array of shape (window_length, len(centers)) containing the extracted windows.
        Invalid windows are filled with NaN.

    Raises
    ------
    ValidationError
        If input parameters are invalid.
    IndexError
        If the calculated indices for any window are out of bounds for the signal array.

    Examples
    --------
    >>> import numpy as np
    >>> signal = np.random.randn(1000)
    >>> centers = np.array([100, 200, 300])
    >>> windows = extract_event_windows(signal, centers, start_offset=50, window_length=100)
    >>> print(f"Windows shape: {windows.shape}")  # (100, 3)
    """
    if not isinstance(signal, np.ndarray) or signal.ndim != 1:
        raise ValidationError(
            "signal must be a 1D numpy array",
            "signal_ndim",
            signal.ndim if hasattr(signal, "ndim") else None,
        )

    if not isinstance(centers, np.ndarray) or centers.ndim != 1:
        raise ValidationError(
            "centers must be a 1D numpy array",
            "centers_ndim",
            centers.ndim if hasattr(centers, "ndim") else None,
        )

    if not isinstance(start_offset, int):
        raise ValidationError(
            "start_offset must be an integer", "start_offset_type", type(start_offset)
        )

    if not isinstance(window_length, int) or window_length <= 0:
        raise ValidationError(
            "window_length must be a positive integer", "window_length", window_length
        )

    if len(centers) == 0:
        logger.warning("Empty centers array provided")
        return np.empty((window_length, 0))

    event_windows = np.full((window_length, len(centers)), np.nan)

    for i, center in enumerate(centers):
        try:
            start_idx = int(np.round(center - start_offset))
            end_idx = start_idx + window_length
            idx = np.arange(start_idx, end_idx)

            if np.any(idx < 0) or np.any(idx >= len(signal)):
                logger.warning(
                    f"Window {i} (center={center}) out of bounds: "
                    f"indices {idx[0]}-{idx[-1]} for signal length {len(signal)}"
                )
                continue

            event_windows[:, i] = signal[idx]

        except Exception as e:
            logger.error(f"Failed to extract window {i} (center={center}): {e}")
            continue

    valid_windows = np.sum(~np.isnan(event_windows[0, :]))
    logger.info(f"Successfully extracted {valid_windows}/{len(centers)} windows")

    return event_windows


def compute_event_statistics(
    event_data: np.ndarray, model_order: int, epsilon: float = 1e-5
) -> Dict[str, Union[np.ndarray, Dict]]:
    """
    Compute conditional statistics for VAR time series events.

    This function computes mean, covariance, and OLS coefficients for VAR time series
    events. It handles both homogeneous and inhomogeneous VAR models.

    Parameters
    ----------
    event_data : np.ndarray
        VAR time series events of shape (nvar * (model_order + 1), time_points, trials).
    model_order : int
        The model order for the VAR process.
    epsilon : float, optional
        Small value for regularization if the matrix is singular. Default is 1e-4.

    Returns
    -------
    Dict[str, Union[np.ndarray, Dict]]
        Dictionary containing the conditional statistics:
            - 'mean': Mean of the events (shape: (nvar * (model_order + 1), time_points))
            - 'Sigma': Covariance matrices
              (shape: (time_points, nvar * (model_order + 1), nvar * (model_order + 1)))
            - 'OLS': Dictionary with:
                - 'At': OLS coefficients (shape: (time_points, nvar, nvar * model_order))
                - 'bt': Residual biases
                - 'Sigma_Et': Residual covariance
                - 'sigma_Et': Residual standard deviation

    Raises
    ------
    ValidationError
        If input parameters are invalid.
    ComputationError
        If computation fails due to numerical issues.

    Examples
    --------
    >>> import numpy as np
    >>> event_data = np.random.randn(6, 50, 10)  # (nvar * (morder + 1), time, trials)
    >>> stats = compute_conditional_event_statistics(event_data, model_order=2)
    >>> print(f"Mean shape: {stats['mean'].shape}")
    >>> print(f"Sigma shape: {stats['Sigma'].shape}")
    """
    if not isinstance(event_data, np.ndarray) or event_data.ndim != 3:
        raise ValidationError(
            "event_data must be a 3D numpy array",
            "event_data_ndim",
            event_data.ndim if hasattr(event_data, "ndim") else None,
        )

    if not isinstance(model_order, int) or model_order <= 0:
        raise ValidationError(
            "model_order must be a positive integer", "model_order", model_order
        )

    if not isinstance(epsilon, (int, float)) or epsilon <= 0:
        raise ValidationError("epsilon must be a positive number", "epsilon", epsilon)

    nvar = event_data.shape[0] // (model_order + 1)
    if event_data.shape[0] != nvar * (model_order + 1):
        raise ValidationError(
            f"event_data shape {event_data.shape[0]} is not compatible "
            f"with model_order {model_order}",
            "event_data_shape",
            event_data.shape,
        )

    if event_data.shape[2] < 2:
        raise ValidationError(
            "At least 2 trials required for statistics computation",
            "n_trials",
            event_data.shape[2],
        )

    try:
        stats = {
            "mean": np.mean(event_data, axis=2),
            "Sigma": np.zeros(
                (
                    event_data.shape[1],
                    nvar * (model_order + 1),
                    nvar * (model_order + 1),
                )
            ),
            "OLS": {"At": np.zeros((event_data.shape[1], nvar, nvar * model_order))},
        }

        for t in range(event_data.shape[1]):
            try:
                temp = event_data[:, t, :] - stats["mean"][:, t, np.newaxis]
                stats["Sigma"][t, :, :] = np.dot(temp, temp.T) / event_data.shape[2]

                Sigma_sub_matrix = stats["Sigma"][t, :nvar, nvar:]
                Sigma_past = stats["Sigma"][t, nvar:, nvar:]

                if np.linalg.det(Sigma_past) > epsilon:
                    Sigma_past_inv = np.linalg.inv(Sigma_past)
                    stats["OLS"]["At"][t, :, :] = np.dot(
                        Sigma_sub_matrix, Sigma_past_inv
                    )
                else:
                    logger.warning(
                        f"Singular Sigma_past at time {t}, using regularization"
                    )
                    Sigma_past_regularized = regularize_if_singular(
                        Sigma_past, epsilon=epsilon
                    )
                    stats["OLS"]["At"][t, :, :] = np.linalg.solve(
                        Sigma_past_regularized, Sigma_sub_matrix.T
                    ).T

            except Exception as e:
                logger.error(f"Failed to compute statistics at time {t}: {e}")
                raise ComputationError(
                    f"Statistics computation failed at time {t}",
                    "statistics_computation",
                    (t,),
                )

        stats["OLS"]["bt"], stats["OLS"]["Sigma_Et"], stats["OLS"]["sigma_Et"] = (
            estimate_residuals(stats)
        )

        logger.info(
            f"Successfully computed statistics for {event_data.shape[1]} time points"
        )
        return stats

    except Exception as e:
        logger.error(f"Statistics computation failed: {e}")
        raise ComputationError(
            f"Statistics computation failed: {e}", "statistics_computation", ()
        )


def extract_event_snapshots(
    time_series: np.ndarray,
    locations: np.ndarray,
    model_order: int,
    lag_step: int,
    start_offset: int,
    extract_length: int,
) -> np.ndarray:
    """
    Extract event snapshots from time series data.

    This function extracts fixed-length snapshots from multivariate time series
    data around specified event locations. Each snapshot includes the current
    time point and lagged data according to the model order.

    Parameters
    ----------
    time_series : np.ndarray
        Time series data of shape (n_vars, n_time_points).
    locations : np.ndarray
        1D array of event location indices.
    model_order : int
        Model order (number of lags to include).
    lag_step : int
        Step size between lags.
    start_offset : int
        Offset from event location to start extraction.
    extract_length : int
        Length of each snapshot to extract.

    Returns
    -------
    np.ndarray
        3D array of shape (n_vars * (model_order + 1), extract_length, n_events)
        containing the extracted snapshots.

    Raises
    ------
    ValidationError
        If input parameters are invalid.
    IndexError
        If any event location is out of bounds.

    Examples
    --------
    >>> import numpy as np
    >>> time_series = np.random.randn(2, 1000)  # (n_vars, time)
    >>> locations = np.array([100, 200, 300])
    >>> snapshots = extract_event_snapshots(
    ...     time_series, locations, model_order=4, lag_step=1,
    ...     start_offset=50, extract_length=100
    ... )
    >>> print(f"Snapshots shape: {snapshots.shape}")  # (10, 100, 3)
    """
    if not isinstance(time_series, np.ndarray) or time_series.ndim != 2:
        raise ValidationError(
            "time_series must be a 2D numpy array",
            "time_series_ndim",
            time_series.ndim if hasattr(time_series, "ndim") else None,
        )

    if not isinstance(locations, np.ndarray) or locations.ndim != 1:
        raise ValidationError(
            "locations must be a 1D numpy array",
            "locations_ndim",
            locations.ndim if hasattr(locations, "ndim") else None,
        )

    if not isinstance(model_order, int) or model_order <= 0:
        raise ValidationError(
            "model_order must be a positive integer", "model_order", model_order
        )

    if not isinstance(lag_step, int) or lag_step <= 0:
        raise ValidationError(
            "lag_step must be a positive integer", "lag_step", lag_step
        )

    if not isinstance(start_offset, int):
        raise ValidationError(
            "start_offset must be an integer", "start_offset", start_offset
        )

    if not isinstance(extract_length, int) or extract_length <= 0:
        raise ValidationError(
            "extract_length must be a positive integer",
            "extract_length",
            extract_length,
        )

    n_vars, n_time_points = time_series.shape

    if len(locations) == 0:
        logger.warning("Empty locations array provided")
        return np.empty((n_vars * (model_order + 1), extract_length, 0))

    min_location = np.min(locations)
    max_location = np.max(locations)

    if min_location < 0 or max_location >= n_time_points:
        raise IndexError(
            f"Event locations {min_location}-{max_location} out of bounds "
            f"for time series length {n_time_points}"
        )

    try:
        n_events = len(locations)
        snapshots = np.zeros((n_vars * (model_order + 1), extract_length, n_events))

        idx1 = np.arange(n_vars * (model_order + 1))
        idx2 = np.tile(np.arange(n_vars), model_order + 1)
        delay = np.tile(np.arange(0, model_order + 1) * lag_step, (n_vars, 1)).flatten()

        for n in range(len(idx1)):
            snapshots[idx1[n], :, :] = extract_event_windows(
                time_series[idx2[n], :],
                locations - delay[n],
                start_offset,
                extract_length,
            )
        valid_snapshots = np.sum(~np.isnan(snapshots[0, 0, :]))
        logger.info(f"Successfully extracted {valid_snapshots}/{n_events} snapshots")

        return snapshots

    except Exception as e:
        logger.error(f"Snapshot extraction failed: {e}")
        raise ComputationError(
            f"Snapshot extraction failed: {e}", "snapshot_extraction", None
        )


def perform_desnap_analysis(inputs: DeSnapParams) -> Dict[str, Union[np.ndarray, Dict]]:
    """
    Perform DeSnap analysis for event-based causality.

    This function implements the DeSnap (De-Snapshot) analysis method for
    event-based causality analysis. It performs a "de-snapshotting" analysis
    to derive unconditional statistics from conditional statistics by accounting
    for a conditioning variable 'D'.

    Parameters
    ----------
    inputs : DeSnapParams
        Configuration object containing all parameters for DeSnap analysis.

    Returns
    -------
    Dict[str, Union[np.ndarray, Dict]]
        Dictionary containing DeSnap analysis results:
            - 'loc_size': Size of locations per bin of D
            - 'p_t', 'q_t': Coefficients from the first linear regression
            - 'd_bin_bar': Mean D values for each bin
            - 'mean_Yt_cond': Mean Yt events per bin of D
            - 'mu_D': Estimated unconditional mean of D
            - 'event_stats_uncond': Unconditional statistics
            - 'cov_pt': Covariance related to p_t
            - 'c': Covariance adjustment factor

    Raises
    ------
    ValidationError
        If input parameters are invalid.
    ComputationError
        If computation fails due to numerical issues.

    Examples
    --------
    >>> from trancit.config import DeSnapParams
    >>> params = DeSnapParams(...)  # Configure parameters
    >>> results = perform_desnap_analysis(params)
    >>> print(f"Results keys: {list(results.keys())}")
    """
    if not isinstance(inputs, DeSnapParams):
        raise ValidationError(
            "inputs must be a DeSnapParams object", "inputs_type", type(inputs)
        )

    try:
        if inputs.d0_max is None and inputs.maxStdRatio is None:
            raise ValidationError(
                "Either d0_max or maxStdRatio must be provided", "d0_max_missing", None
            )

        if inputs.N_d <= 0:
            raise ValidationError(
                "N_d (number of bins) must be positive", "N_d", inputs.N_d
            )

        if inputs.detection_signal.ndim != 1:
            raise ValidationError(
                "detection_signal must be 1D",
                "detection_signal_ndim",
                inputs.detection_signal.ndim,
            )

        if inputs.original_signal.ndim != 2:
            raise ValidationError(
                "original_signal must be 2D",
                "original_signal_ndim",
                inputs.original_signal.ndim,
            )

        if inputs.detection_signal.shape[0] != inputs.original_signal.shape[1]:
            raise ValidationError(
                "detection_signal and original_signal must have same time dimension",
                "signal_lengths",
                (inputs.detection_signal.shape[0], inputs.original_signal.shape[1]),
            )

        if inputs.d0_max is None:
            if inputs.maxStdRatio is not None:
                inputs.d0_max = np.mean(
                    inputs.detection_signal
                ) + inputs.maxStdRatio * np.std(inputs.detection_signal)
            else:
                inputs.d0_max = np.max(inputs.detection_signal)

        logger.info(
            f"DeSnap analysis: d0={inputs.d0}, d0_max={inputs.d0_max}, N_d={inputs.N_d}"
        )

        bin_step = abs(inputs.d0_max - inputs.d0) / inputs.N_d
        d_bin_edges = np.arange(inputs.d0, inputs.d0_max + bin_step + 1e-12, bin_step)
        num_bins = len(d_bin_edges)

        d_bin_mean_detection = np.full(num_bins, np.nan)
        num_input_channels = inputs.original_signal.shape[0]
        num_snapshot_vars = num_input_channels * (inputs.morder + 1)
        mean_events_cond_binned = np.full(
            (num_bins, num_snapshot_vars, inputs.l_extract), np.nan
        )

        DeSnap_results = {
            "loc_size": np.full(num_bins, np.nan),
            "event_stats": inputs.event_stats,
        }

        logger.info("Processing bins of conditioning variable D ...")
        current_bin_uplim = np.max(inputs.detection_signal)

        for n, current_bin_lolim in enumerate(d_bin_edges):
            mask = (inputs.detection_signal >= current_bin_lolim) & (
                inputs.detection_signal < current_bin_uplim
            )

            d_bin_mean_detection[n] = np.mean(inputs.detection_signal[mask])
            temp_loc = np.where(mask)[0]

            valid_locs = temp_loc.copy()
            # valid_locs = valid_locs[
            #     inputs.original_signal.shape[0] - valid_locs >= inputs.l_extract - inputs.l_start
            # ]
            # valid_locs = valid_locs[
            #     valid_locs - inputs.l_start - (inputs.morder * inputs.tau) >= 0
            # ]

            DeSnap_results["loc_size"][n] = len(valid_locs)

            if len(valid_locs) > 0:
                try:
                    events_binned = extract_event_snapshots(
                        inputs.original_signal,
                        valid_locs,
                        inputs.morder,
                        inputs.tau,
                        inputs.l_start,
                        inputs.l_extract,
                    )
                    if events_binned.shape[2] > 0:
                        mean_events_cond_binned[n, :, :] = np.mean(
                            events_binned, axis=2
                        )
                    else:
                        logger.warning(
                            f"No snapshots extracted for bin {n+1} despite "
                            f"{len(valid_locs)} valid_locs"
                        )
                except Exception as e:
                    logger.error(f"Failed to extract snapshots for bin {n}: {e}")
            else:
                logger.warning(
                    f"No valid locations for snapshot extraction in bin {n+1}"
                )

        # First Linear Regression: Fit p_t and q_t
        logger.info("Performing first linear regression for p_t and q_t...")
        try:
            from .helpers import compute_multi_variable_linear_regression

            p_t, q_t = compute_multi_variable_linear_regression(
                d_bin_mean_detection, mean_events_cond_binned
            )
            DeSnap_results["p_t"] = p_t
            DeSnap_results["q_t"] = q_t
            DeSnap_results["d_bin_bar"] = d_bin_mean_detection
            DeSnap_results["mean_Yt_cond"] = mean_events_cond_binned
        except Exception as e:
            logger.error(f"First linear regression failed: {e}")
            raise ComputationError(
                f"First linear regression failed: {e}", "linear_regression_1", None
            )

        # Second Linear Regression: Estimate mu_D (unconditional mean of D)
        p_t_flat = p_t.reshape(-1, 1)
        q_t_flat = -q_t.reshape(-1)

        nan_mask_regression2 = ~np.isnan(p_t_flat.ravel()) & ~np.isnan(q_t_flat)
        if not np.any(nan_mask_regression2):
            raise ComputationError(
                "All p_t or q_t values are NaN, cannot compute mu_D",
                "mu_D_computation",
                None,
            )

        p_t_flat_valid = p_t_flat[nan_mask_regression2]
        q_t_flat_valid = q_t_flat[nan_mask_regression2]

        try:
            DeSnap_results["mu_D"] = np.linalg.lstsq(
                p_t_flat_valid, q_t_flat_valid, rcond=None
            )[0][0]
        except Exception as e:
            logger.error(f"Second linear regression failed: {e}")
            raise ComputationError(
                f"Second linear regression failed: {e}", "linear_regression_2", None
            )

        DeSnap_results["event_stats_uncond"] = {}
        DeSnap_results["event_stats_uncond"]["mean"] = (
            q_t + p_t * DeSnap_results["mu_D"]
        )

        # Third Linear Regression: Compute Covariance Adjustment Factor 'c'
        logger.info(
            "Performing third linear regression for covariance adjustment factor 'c'..."
        )
        DeSnap_results["cov_pt"] = np.full(
            (inputs.l_extract, num_snapshot_vars, num_snapshot_vars), np.nan
        )
        for t in range(inputs.l_extract):
            DeSnap_results["cov_pt"][t, :, :] = np.outer(p_t[:, t], p_t[:, t])

        try:
            if inputs.diff_flag:
                x_reg_c = np.diff(DeSnap_results["cov_pt"], axis=0)
                y_reg_c = np.diff(inputs.event_stats["Sigma"], axis=0)
                DeSnap_results["c"] = np.linalg.lstsq(
                    x_reg_c.reshape(-1, 1), y_reg_c.reshape(-1), rcond=None
                )[0][0]
            else:
                x_reg_c_levels = DeSnap_results["cov_pt"][:, 0, 0]
                y_reg_c_levels = inputs.event_stats["Sigma"][:, 0, 0]
                X_design_c = np.vstack([np.ones_like(x_reg_c_levels), x_reg_c_levels]).T
                temp_coeffs_c = np.linalg.lstsq(X_design_c, y_reg_c_levels, rcond=None)[
                    0
                ]
                DeSnap_results["c"] = temp_coeffs_c[1]
        except Exception as e:
            logger.error(f"Third linear regression failed: {e}")
            raise ComputationError(
                f"Third linear regression failed: {e}", "linear_regression_3", None
            )

        DeSnap_results["event_stats_uncond"]["Sigma"] = (
            inputs.event_stats["Sigma"] - DeSnap_results["c"] * DeSnap_results["cov_pt"]
        )

        logger.info("Calculating unconditional AR coefficients...")
        try:
            nvar_actual = inputs.event_stats["OLS"]["At"].shape[1]
        except (KeyError, AttributeError, IndexError):
            raise ValidationError(
                "Could not determine 'nvar_actual' from inputs.event_stats['OLS']['At']",
                "ols_at_structure",
                None,
            )

        DeSnap_results["event_stats_uncond"]["OLS"] = {}
        DeSnap_results["event_stats_uncond"]["OLS"]["At"] = np.full(
            (inputs.l_extract, nvar_actual, nvar_actual * inputs.morder), np.nan
        )

        for t in range(inputs.l_extract):
            try:
                Sigma_yx_uncond = DeSnap_results["event_stats_uncond"]["Sigma"][
                    t, :nvar_actual, nvar_actual:
                ]
                Sigma_xx_uncond = DeSnap_results["event_stats_uncond"]["Sigma"][
                    t, nvar_actual:, nvar_actual:
                ]

                Sigma_xx_uncond_reg = regularize_if_singular(Sigma_xx_uncond)
                if not np.allclose(Sigma_xx_uncond, Sigma_xx_uncond_reg):
                    logger.warning(
                        f"DeSnap: Applied regularization to Sigma_xx_uncond at time step {t}"
                    )

                DeSnap_results["event_stats_uncond"]["OLS"]["At"][t, :, :] = (
                    Sigma_yx_uncond @ np.linalg.inv(Sigma_xx_uncond_reg)
                )
            except np.linalg.LinAlgError:
                logger.warning(
                    f"DeSnap: Singular matrix at time step {t}, using pseudo-inverse"
                )
                DeSnap_results["event_stats_uncond"]["OLS"]["At"][t, :, :] = (
                    Sigma_yx_uncond @ np.linalg.pinv(Sigma_xx_uncond_reg)
                )
            except Exception as e:
                logger.error(f"Failed to compute AR coefficients at time step {t}: {e}")
                continue

        logger.info(f"Successfully completed DeSnap analysis with {num_bins} bins")
        return DeSnap_results

    except Exception as e:
        logger.error(f"DeSnap analysis failed: {e}")
        raise ComputationError(f"DeSnap analysis failed: {e}", "desnap_analysis", ())



"""
Models module for TranCIT: Transient Causal Interaction.

This module provides implementations for:
- VAR model estimation: Vector Autoregressive model fitting
- BIC model selection: Bayesian Information Criterion for model order selection
- Model validation: Diagnostics and validation of fitted models
- Model selection: High-level model selection utilities

All model estimators provide consistent interfaces for time series modeling
and support both homogeneous and inhomogeneous VAR models.

Example
-------
>>> import numpy as np
>>> from trancit.models import VAREstimator, BICSelector, select_model_order
>>>
>>> # Create VAR estimator
>>> estimator = VAREstimator(model_order=4, time_mode="inhomo")
>>>
>>> # Generate sample data
>>> data = np.random.randn(2, 1000, 20)  # (n_vars, n_obs, n_trials)
>>>
>>> # Estimate VAR coefficients
>>> coefficients, residuals, log_likelihood, hessian_sum = estimator.estimate_var_coefficients(
...     data, max_model_order=4
... )
>>>
>>> # Select optimal model order
>>> bic_scores, optimal_orders, log_likelihoods, penalty_terms = select_model_order(
...     data, max_model_order=6, time_mode="inhomo"
... )
>>>
>>> print(f"Coefficients shape: {coefficients.shape}")
>>> print(f"Optimal model orders: {optimal_orders}")
"""

from .bic_selection import BICSelector
from .model_selection import select_model_order
from .model_validation import ModelValidator
from .var_estimation import VAREstimator

__all__ = [
    "VAREstimator",
    "BICSelector",
    "ModelValidator",
    "select_model_order",
]
